#####################################################
###   A registry of callable view functions
###   in the DMP system

from django.apps import apps, AppConfig
from django.core.exceptions import ImproperlyConfigured

from .util import get_dmp_instance, DMP_OPTIONS

import threading


# lock to keep register() thread safe
rlock = threading.RLock()

# the DMP-enabled AppConfigs in the system (by name)
DMP_ENABLED_APPS = {}


###############################################################
###   Registration functions for apps and view functions
###

def get_dmp_apps():
    '''Returns a sequence of DMP-enabled apps'''
    return DMP_ENABLED_APPS.values()


def is_dmp_app(app):
    '''
    Returns True if the given app is a DMP-enabled app.  The app parameter can
    be either the name of the app or an AppConfig object.
    '''
    if app is None:
        return False
    if isinstance(app, AppConfig):
        app = app.name
    return app in DMP_ENABLED_APPS


def ensure_dmp_app(app):
    '''Raises an AssertionException with a help mesage if the given app is not registered as a DMP app'''
    if not is_dmp_app(app):
        raise ImproperlyConfigured('''DMP's app-specific render functions were not placed on the request object. Likely reasons:
1) A template was rendered before the middleware's process_view() method was called (move after middleware call or use DMP's render_template() function which can be used anytime).
2) This app is not registered as a Django app (ensure the app is listed in `INSTALLED_APPS` in settings file).
3) This app is not registered as a DMP-enabled app (check `APP_DISCOVERY` in settings file or see DMP docs).''')



def register_dmp_app(app, inject_urls=False):
    '''
    Registers an app as a "DMP-enabled" app.  Normally, DMP does this
    automatically during system startup.

    If auto-discovery is disabled, call this method to register
    an app as DMP-enabled.  This should be done in the app's
    AppConfig.ready() method so registration happens *before* Django
    processes through urls.py.
    '''
    # this can only be done *before* django processes through urls.py
    if apps.ready:
        raise ImproperlyConfigured("App registration attempted too late in the startup process. "
                                   "Apps can no longer be registered with DMP because Django has already set up. "
                                   "Please call register_dmp_app() earlier in the process (ideally during AppConfig.ready() in yourapp/apps.py).")

    # since this only runs at startup, this lock doesn't affect performance
    if isinstance(app, str):
        app = apps.get_app_config(app)
    with rlock:
        # first time for this app, so add to our dictionary
        DMP_ENABLED_APPS[app.name] = app

        # set up the template, script, and style renderers
        # these create and cache just by accessing them
        get_dmp_instance().get_template_loader(app, 'templates', create=True)
        get_dmp_instance().get_template_loader(app, 'scripts', create=True)
        get_dmp_instance().get_template_loader(app, 'styles', create=True)




from .util import log, DMP_OPTIONS

import inspect
from collections import defaultdict


################################################################
###  An abstract pass-through decorator.

_UNDEFINED_FUNCTION = object()

class PassthroughDecorator(object):
    '''
    A pass-through decorator that can be called with @decorator or @decorator(a=1, b=2).
    The decorate() method is called to perform the behavior of the decorator on the function.
    This class instance does not stay in the decorator call chain.  The primary use case
    of this decorator is place an attribute on the function.
    
    @wraps is not used with this because it is a pass-through.  The decorator disappears
    after doing its initial work and does not stay within the call chain.

    Note that it is not possible to use this decorator with the first argument being a callable
    because Python will take it as the decorated function.
    '''
    def __new__(cls, func=_UNDEFINED_FUNCTION, *args, **kwargs):
        # if we have a function, it was called without arguments: @decorator
        # annotate and return the function (an instance never gets created)
        if func is not _UNDEFINED_FUNCTION and callable(func) and len(args) == 0 and len(kwargs) == 0:
            cls.decorate(func)
            return func

        # if we don't have a function, it was specified with arguments: @decorator(a=1, b=2)
        # we need to make an object, let __init__ run, and let python call __call__.
        return super().__new__(cls)

    def __init__(self, func=None, *args, **kwargs):
        self.args = (func, ) + args
        self.kwargs = kwargs

    def __call__(self, func):
        self.decorate(func, *self.args, **self.kwargs)
        return func

    @classmethod
    def decorate(cls, func, *args, **kwargs):
        '''Subclasses should override this method to make the annotation'''
        raise NotImplementedError('The `annotate` method must be implemented by subclasses.')



###################################################################
###  A pass-through decorator that annotates with the
###  decorator args and kwargs.

ANNOTATION_DECORATOR_KEY = '_dmp_annotation_decorator'


class KeywordArgDecorator(PassthroughDecorator):
    '''
    Caches the kwargs given in annotate on the annotated function.
    Multiple decorators of the same type can be placed on the same function.

    Unless force=True, raises NotDecoratedError if cls is not decorating the function.
    '''
    DEFAULT_KWARGS = {}

    @classmethod
    def _get_list_for_class(cls, func, force=False):
        func = inspect.unwrap(func, stop=lambda f: hasattr(f, ANNOTATION_DECORATOR_KEY))  # try to get the actual function
        if not hasattr(func, ANNOTATION_DECORATOR_KEY):
            if not force:
                raise NotDecoratedError('The function is not decorated with decorator: {}'.format(cls.__qualname__))
            setattr(func, ANNOTATION_DECORATOR_KEY, defaultdict(list))
        return getattr(func, ANNOTATION_DECORATOR_KEY)[cls]

    @classmethod
    def decorate(cls, func, *args, **kwargs):
        '''Does the work of the decorator: caches the kwargs for later retrieval.'''
        updated_kwargs = cls.DEFAULT_KWARGS.copy()
        updated_kwargs.update(kwargs)
        # add the ANNOTATION_DECORATOR_KEY to the function
        cls._get_list_for_class(func, force=True).append(updated_kwargs)

    @classmethod
    def get_kwargs(cls, func):
        '''
        Returns a list of cached kwargs for this decorator type on the given function.
        It is a list because the decorator could be set more than once on a given function.
        Raises NotDecoratedError if the decorator is not on the function.
        The list will always have at least one element (or NotDecoratedError would have raised).
        '''
        return cls._get_list_for_class(func)

    @classmethod
    def is_decorated(cls, func):
        '''Returns whether the function is decorated with this decorator type'''
        try:
            cls._get_list_for_class(func)
        except NotDecoratedError:
            return False
        return True


class NotDecoratedError(Exception):
    pass



##############################################################
###   Decorators for view functions

class view_function(KeywordArgDecorator):
    '''
    A decorator to signify which view functions are "callable" by web browsers.

    Any endpoint function, such as process_request, must be decorated to be callable:

        @view_function
        function process_request(request):
            ...

    Class-based views don't need to be decorated because we allow anything that extends Django's View class.
    '''
    DEFAULT_KWARGS = {
        'converter': None,
    }




from django.apps import apps
from django.template import Context

import mako.runtime

from .runner import ProviderRun
from ..util import get_dmp_instance

import io



#########################################################
###  Primary functions

def links(tself, version_id=None, group=None):
    '''Returns the HTML for the given provider group (or all groups if None)'''
    provider_run = ProviderRun(tself, version_id, group)
    return provider_run.get_content()


def template_links(request, app, template_name, context=None, version_id=None, group=None, force=True):
    '''
    Returns the HTML for the given provider group, using an app and template name.
    This method should not normally be used (use links() instead).  The use of
    this method is when provider need to be called from regular python code instead
    of from within a rendering template environment.
    '''
    if isinstance(app, str):
        app = apps.get_app_config(app)
    if context is None:
        context = {}
    template_obj = get_dmp_instance().get_template_loader(app, create=True).get_mako_template(template_name, force=force)
    return template_obj_links(request, template_obj, context, version_id, group)


def template_obj_links(request, template_obj, context=None, version_id=None, group=None):
    '''
    Returns the HTML for the given provider group, using a template object.
    This method should not normally be used (use links() instead).  The use of
    this method is when provider need to be called from regular python code instead
    of from within a rendering template environment.
    '''
    # the template_obj can be a MakoTemplateAdapter or a Mako Template
    # if our DMP-defined MakoTemplateAdapter, switch to the embedded Mako Template
    template_obj = getattr(template_obj, 'mako_template', template_obj)
    # create a mako context so it seems like we are inside a render
    context_dict = {
        'request': request,
    }
    if isinstance(context, Context):
        for d in context:
            context_dict.update(d)
    elif context is not None:
        context_dict.update(context)
    mako_context = create_mako_context(template_obj, **context_dict)
    return links(mako_context['self'], version_id=version_id, group=group)


def create_mako_context(template_obj, **kwargs):
    # I'm hacking into private Mako methods here, but I can't see another
    # way to do this.  Hopefully this can be rectified at some point.
    kwargs.pop('self', None)  # some contexts have self in them, and it messes up render_unicode below because we get two selfs
    runtime_context = mako.runtime.Context(io.BytesIO(), **kwargs)
    runtime_context._set_with_template(template_obj)
    _, mako_context = mako.runtime._populate_self_namespace(runtime_context, template_obj)
    return mako_context







from django.core.management.commands.makemessages import Command as MakeMessagesCommand

from django_mako_plus.util import get_dmp_instance
from django_mako_plus.registry import get_dmp_apps

import os
import os.path


class Command(MakeMessagesCommand):
    help = "Makes messages for Mako templates.  The native makemessages in Django doesn't understand Mako files, so this command compiles all your templates so it can find the compiled .py versions of the templates."


    def add_arguments(self, parser):
        # must call super because not directly subclassing BaseCommand
        super(Command, self).add_arguments(parser)
        # add our additional command
        parser.add_argument('--extra-gettext-option', default=[], dest='extra_gettext_option', action='append', help="Add an additional option to be passed to gettext. Ex: --extra-gettext-option='--keyword=mytrans'")


    def handle(self, *args, **options):
        # go through each dmp_enabled app and compile its mako templates
        for app_config in get_dmp_apps():
            self.compile_mako_files(app_config)

        # add any extra xgettext_options (the regular makemessages doesn't do this, and I need to include other aliases like _(), _z(), etc.
        for opt in options.get('extra_gettext_option', []):
            self.xgettext_options.append(opt)

        # call the superclass command
        return super(Command, self).handle(*args, **options)


    def compile_mako_files(self, app_config):
        '''Compiles the Mako templates within the apps of this system'''
        # go through the files in the templates, scripts, and styles directories
        moddir = os.path.dirname(app_config.module.__file__)
        for subdir_name in ('templates', 'scripts', 'styles'):
            subdir = os.path.join(moddir, subdir_name)
            if os.path.exists(subdir):
                for filename in os.listdir(subdir):
                    fname, ext = os.path.splitext(filename)
                    if ext.lower() in ( '.htm', '.html', '.jsm', '.cssm' ):
                        # create the template object, which creates the compiled .py file
                        renderer = get_dmp_instance().get_template_loader(app_config.name, subdir_name)
                        renderer.get_template(filename)



from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from mako.exceptions import MakoException
from django_mako_plus.registry import get_dmp_apps
from django_mako_plus.provider import create_mako_context
from django_mako_plus.provider.runner import ProviderRun, create_factories
from django_mako_plus.util import get_dmp_instance, split_app, DMP_OPTIONS

import glob
import os, os.path, shutil
import json
from collections import OrderedDict



class Command(BaseCommand):
    args = ''
    help = 'Removes compiled template cache folders in your DMP-enabled app directories.'
    can_import_settings = True


    def add_arguments(self, parser):
        parser.add_argument(
            '--verbose',
            action='store_true',
            dest='verbose',
            default=False,
            help='Set verbosity to level 3 (see --verbosity).'
        )
        parser.add_argument(
            '--quiet',
            action='store_true',
            dest='quiet',
            default=False,
            help='Set verbosity to level 0, which silences all messages (see --verbosity).'
        )
        parser.add_argument(
            'appname',
            type=str,
            nargs='*',
            help='The name of an app. If omitted, __entry__.js files are created in all DMP apps.'
        )
        parser.add_argument(
            '--overwrite',
            action='store_true',
            dest='overwrite',
            default=False,
            help='Overwrite existing __entry__.js if necessary.'
        )


    def handle(self, *args, **options):
        # save the options for later
        self.options = options
        if self.options['verbose']:
            self.options['verbosity'] = 3
        if self.options['quiet']:
            self.options['verbosity'] = 0

        # ensure we have a base directory
        try:
            if not os.path.isdir(os.path.abspath(settings.BASE_DIR)):
                raise CommandError('Your settings.py BASE_DIR setting is not a valid directory.  Please check your settings.py file for the BASE_DIR variable.')
        except AttributeError as e:
            print(e)
            raise CommandError('Your settings.py file is missing the BASE_DIR setting.')

        # run for each dmp-enabled app
        self.factories = create_factories('WEBPACK_PROVIDERS')
        for config in get_dmp_apps():
            if not options['appname'] or config.name in options['appname']:
                self.create_entry(config)


    def message(self, msg, level=1):
        '''Print a message to the console'''
        # verbosity=1 is the default if not specified in the options
        if self.options['verbosity'] >= level:
            print(msg)


    def create_entry(self, config):
        '''Creates a webpack __entry__.js file in the given app'''
        self.message('Searching {}/'.format(config.name))
        entry_filename = os.path.join(config.path, 'scripts', '__entry__.js')
        script_map = self.generate_script_map(config)
        # write the file
        if not self.options['overwrite'] and os.path.exists(entry_filename):
            raise ValueError('\tRefusing to destroy existing file: {} (use --overwrite option or remove the file)'.format(entry_filename), 1)
        if len(script_map) == 0:
            self.message('\tno matching scripts')
        else:
            self.message('\twriting {} matching templates in {}'.format(len(script_map), os.path.relpath(entry_filename, settings.BASE_DIR)))
            with open(entry_filename, 'w') as fout:
                fout.write('(context => {\n')
                for page, scripts in script_map.items():
                    fout.write('    DMP_CONTEXT.appBundles["%s"] = () => { %s; };\n' % (
                        page,
                        '; '.join([ 'require("%s")' % (s,) for s in scripts ]),
                    ))
                fout.write('})(DMP_CONTEXT.get());\n')


    def generate_script_map(self, config):
        '''
        Maps templates in this app to their scripts.  This function iterates through
        app/templates/* to find the templates in this app.  Returns the following
        dictionary:

        {
            'app/template1': [ './scripts/template1.js', './scripts/supertemplate1.js' ],
            'app/template2': [ './scripts/template2.js', './scripts/supertemplate2.js', './scripts/supersuper2.js' ],
            ...
        }

        Any files or subdirectories starting with double-underscores (e.g. __dmpcache__) are skipped.
        '''
        script_map = OrderedDict()
        template_root = os.path.join(config.path, 'templates')
        def recurse(folder):
            for filename in os.listdir(os.path.join(template_root, folder)):
                if filename.startswith('__'):
                    continue
                filerel = os.path.join(folder, filename)
                filepath = os.path.join(os.path.join(template_root, filerel))
                if os.path.isdir(filepath):
                    recurse(filerel)

                elif os.path.isfile(filepath):
                    scripts = self.template_scripts(config, filerel)
                    key = '{}/{}'.format(config.name, os.path.splitext(filerel)[0])
                    if len(scripts) > 0:
                        script_map[key] = scripts
                        self.message('\t{}: {}'.format(key, scripts), 3)
                    else:
                        self.message('\t{}: none found'.format(key), 3)

        recurse('')
        return script_map


    def template_scripts(self, config, template_name):
        '''
        Returns a list of scripts used by the given template object AND its ancestors.

        This runs a ProviderRun on the given template (as if it were being displayed).
        This allows the WEBPACK_PROVIDERS to provide the JS files to us.

        For this algorithm to work, providers must populate the provider data dictionary like this example.
        the built-in JsLinkProvider does this already.
        provider_data = {
            'urls': [
                '/static/app/scripts/first.js',
                '/static/app/scripts/second.js',
                ...
            ]
        }
        '''
        template_obj = get_dmp_instance().get_template_loader(config, create=True).get_mako_template(template_name, force=True)
        mako_context = create_mako_context(template_obj)
        inner_run = ProviderRun(mako_context['self'], factories=self.factories)
        inner_run.get_content()
        scripts = []
        for data in inner_run.provider_data:
            scripts_dir = os.path.join(config.name, 'scripts')
            for url in data.get('urls', []):
                url = url.split('?', 1)[0]
                scripts.append(os.path.join('.', os.path.relpath(url, scripts_dir)))
        return scripts








from django.apps import apps
from django.db.models import Model, ObjectDoesNotExist
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest

from .exceptions import RedirectException
from .util import DMP_OPTIONS, log

import inspect, datetime, decimal, sys
from collections import namedtuple
from operator import attrgetter


##################################################################
###   ConversionTask object

class ConversionTask(object):
    '''
    A (mostly) data class that holds meta-information about a conversion
    task.  This object is sent into each converter function.
    '''
    def __init__(self, request, module, function, kwargs):
        self.converter = _check_converter(kwargs.get('converter'))
        self.request = request
        self.module = module
        self.function = function
        self.kwargs = kwargs       # kwargs from the @view_function decorator


##################################################################
###   ConverterInfo class: see @convert_method below

class ConverterInfo(object):
    '''
    Used by @convert_method to track types on a converter method.
    '''
    CONVERT_TYPES_KEY = '_dmp_converter_infos'
    DECLARED_ORDER = 0

    def __init__(self, convert_type, convert_func):
        self.convert_type = convert_type
        self.convert_func = convert_func
        self.sort_key = None
        self.declared_order = ConverterInfo.DECLARED_ORDER  # order it was declared in source code
        ConverterInfo.DECLARED_ORDER = 0 if ConverterInfo.DECLARED_ORDER == sys.maxsize else ConverterInfo.DECLARED_ORDER + 1

    def _delayed_init(self):
        # we allow the type for models to be declared as a string, such as "auth.User"
        # because real types can't be referenced until Django finishes initialization
        # switch any strings over to the real type now
        if isinstance(self.convert_type, str):
            try:
                app_name, model_name = self.convert_type.split('.')
            except ValueError:
                raise ImproperlyConfigured('"{}" is not a valid converter type. String-based converter types must be specified in "app.Model" format.'.format(self.convert_type))
            try:
                self.convert_type = apps.get_model(app_name, model_name)
            except LookupError as e:
                raise ImproperlyConfigured('"{}" is not a valid model name. {}'.format(self.convert_type, e))

        # we reverse sort by ( len(mro), source code order ) so subclasses match first
        # on same types, last declared method sorts first
        self.sort_key = ( -1 * len(inspect.getmro(self.convert_type)), -1 * self.declared_order )


##################################################################
###   Default converter class


class BaseConverter(object):

    @staticmethod
    def convert_method(*class_types):
        '''
        Decorator for converter methods in DefaultConverter.
        '''
        def inner(func):
            setattr(func, ConverterInfo.CONVERT_TYPES_KEY, [ ConverterInfo(ct, func) for ct in class_types ])
            return func
        return inner


    def __init__(self):
        self.converters = []
        self._delayed_init_run = False


    def _delayed_init(self):
        '''
        Populates the registry of types to converters for this Converter object.  This method is
        called each time a request is processed, so if overridden, be sure to short circuit as needed.
        This implementation short circuits until all apps are ready.
        '''
        # This is called by _check_converter() below, and it is delayed (instead of using __init__)
        # because model class strings can't be switched to models until the app registry is populated.
        # This could be called once from DMP's AppConfig.ready(), but since each view function can have
        # a separate converters, and since the default converter can be changed at any time, I'm doing
        # it when a ConversionTask object is created (part of DMP request processing).
        if self._delayed_init_run or not apps.ready:
            return
        self._delayed_init_run = True

        # converter methods within this class (subclass) were tagged earlier by the decorator
        # with the CONVERT_TYPES_KEY.  go through all methods in this class (subclass) and
        # gather the ConverterInfo objects into a big list
        for name, method in inspect.getmembers(self, inspect.ismethod):
            self.converters.extend(getattr(method, ConverterInfo.CONVERT_TYPES_KEY, ()))

        # allow each ConverterInfo to init, such as convert any model string "auth.User" to real class type
        for ci in self.converters:
            ci._delayed_init()

        # sort the most specialized types to be first.  this allows subclass types to match instead
        # of their superclasses (if both are in the list).
        self.converters.sort(key=attrgetter('sort_key'))


    def __call__(self, value, parameter, task):
        '''
        Converts the given value to the type expected by the view function.

            value:      value from request.dmp.urlparams to convert
                        The value will always be a string, even if empty '' (never None).

            parameter:  an instance of ViewParameter.  See that class above for more information.

            task:       meta-information about the current conversion task (as a named tuple):
                        module:             A reference to the module containing the view function.
                        function:           A reference to the view function.
                        converter:          A reference to the currently-running converter callable.
                        urlparams:          The unconverted request.dmp.urlparams list of strings from the url,
                                            provided for cases when converting a value depends on another.

        This method should do one of the following:

            1. Return the converted value (desired type is in parameter.type) to be placed in converted_args.

            2. Raise an exception to signal an error or validation problem, such as:
                    a. ValueError if unconvertable value (caught by DMP and and raises Http404)
                    b. DMP's RedirectException
                    c. DMP's InternalRedirectException
                    d. Django's Http404 exception
                    e. Any other exception (caught by DMP and and raises Http404)
               See the @view_parameter decorator for more information on handling conversion errors.
        '''
        # we don't convert anything without type hints
        if parameter.type is inspect.Parameter.empty:
            return value

        # find the converter method for this type
        # I'm iterating through the list to find the most specific match first
        # The list is sorted by specificity so subclasses come first
        for ci in self.converters:
            if issubclass(parameter.type, ci.convert_type):
                return ci.convert_func.__get__(self, self.__class__)(value, parameter, task)

        # we should never get here because DefaultConverter below has a converter for <object>
        # (unless a project entirely swaps out the DefaultConverter for another subclass)
        raise ValueError('No parameter converter exists for type: {}. Either remove the type hint or subclass the DMP DefaultConverter class.'.format(parameter.type))



class DefaultConverter(BaseConverter):
    '''
    The default converter that comes with DMP.  URL parameter converters
    can be any callable.  This implementation is an extensible class with pluggable
    methods for conversion of different types.

    The processing function is in super: BaseConverter.__call__().
    '''
    def _check_default(self, value, parameter, task, default_chars):
        '''Returns the default if the value is "empty"'''
        # not using a set here because it fails when value is unhashable
        if value in default_chars:
            if parameter.default is inspect.Parameter.empty:
                raise ValueError('Value was empty, but no default value is given in view function for parameter: {} ({})'.format(parameter.position, parameter.name))
            return parameter.default
        return value


    @BaseConverter.convert_method(HttpRequest)
    def convert_http_request(self, value, parameter, task):
        '''
        Pass through for the request object (first parameter in every view call).
        The request is run through the conversion process for consistency in parameter handling,
        but I can't see a reason it would ever need to be "converted" outside of middleware.
        '''
        return value


    @BaseConverter.convert_method(object)
    def convert_object(self, value, parameter, task):
        '''
        Fallback converter when nothing else matches:
            '', None convert to parameter default
            Anything else is returned as-is
        '''
        return self._check_default(value, parameter, task, ( '', None ))


    @BaseConverter.convert_method(str)
    def convert_str(self, value, parameter, task):
        '''
        Converts to string:
            '', None convert to parameter default
            Anything else is returned as-is (url params are already strings)
        '''
        return self._check_default(value, parameter, task, ( '', None ))


    @BaseConverter.convert_method(int, float)
    def convert_number(self, value, parameter, task):
        '''
        Converts to int or float:
            '', '-', None convert to parameter default
            Anything else uses int() or float() constructor
        '''
        value = self._check_default(value, parameter, task, ( '', '-', None ))
        if value is None or isinstance(value, (int, float)):
            return value
        return parameter.type(value)  # int() or float()


    @BaseConverter.convert_method(decimal.Decimal)
    def convert_decimal(self, value, parameter, task):
        '''
        Converts to decimal.Decimal:
            '', '-', None convert to parameter default
            Anything else uses Decimal constructor
        '''
        value = self._check_default(value, parameter, task, ( '', '-', None ))
        if value is None or isinstance(value, decimal.Decimal):
            return value
        return decimal.Decimal(value)


    @BaseConverter.convert_method(bool)
    def convert_boolean(self, value, parameter, task, default=False):
        '''
        Converts to boolean (only the first char of the value is used):
            '', '-', None convert to parameter default
            'f', 'F', '0', False always convert to False
            Anything else converts to True.
        '''
        value = self._check_default(value, parameter, task, ( '', '-', None ))
        if isinstance(value, bool):
            return value
        if isinstance(value, str) and len(value) > 0:
            value = value[0]
        return value not in ( 'f', 'F', '0', False, None )


    @BaseConverter.convert_method(datetime.datetime)
    def convert_datetime(self, value, parameter, task):
        '''
        Converts to datetime.datetime:
            '', '-', None convert to parameter default
            The first matching format in settings.DATETIME_INPUT_FORMATS converts to datetime
        '''
        value = self._check_default(value, parameter, task, ( '', '-', None ))
        if value is None or isinstance(value, datetime.datetime):
            return value
        for fmt in settings.DATETIME_INPUT_FORMATS:
            try:
                return datetime.datetime.strptime(value, fmt)
            except (ValueError, TypeError):
                continue
        raise ValueError("'{}' does not match a format in settings.DATETIME_INPUT_FORMATS".format(value))


    @BaseConverter.convert_method(datetime.date)
    def convert_date(self, value, parameter, task):
        '''
        Converts to datetime.date:
            '', '-', None convert to parameter default
            The first matching format in settings.DATE_INPUT_FORMATS converts to datetime
        '''
        value = self._check_default(value, parameter, task, ( '', '-', None ))
        if value is None or isinstance(value, datetime.date):
            return value
        for fmt in settings.DATE_INPUT_FORMATS:
            try:
                return datetime.datetime.strptime(value, fmt).date()
            except (ValueError, TypeError):
                continue
        raise ValueError("'{}' does not match a format in settings.DATE_INPUT_FORMATS".format(value))


    @BaseConverter.convert_method(Model)  # django models.Model
    def convert_id_to_model(self, value, parameter, task):
        '''
        Converts to a Model object.
            '', '-', '0', None convert to parameter default
            Anything else is assumed an object id and sent to `.get(id=value)`.
        '''
        value = self._check_default(value, parameter, task, ( '', '-', '0', None ))
        if isinstance(value, (int, str)):  # only convert if we have the id
            return parameter.type.objects.get(id=value)
        return value


####################################################
###   Setting and getting of default converter

DEFAULT_CONVERTER_KEY = '_dmp_default_converter'

def _check_converter(converter):
    '''
    Checks the given converter and returns a (potentially) modified converter.
    If None: returns the system default converter.
    If a class: returns an instance of the class.
    '''
    # default or instantiate if necessary
    if converter is None:
        converter = get_default_converter()
    elif inspect.isclass(converter):
        converter = converter()
    # ensure callable
    if not callable(converter):
        raise ValueError('Converters must be callable or classes that implements the __call__ method.')
    # if a subclass of BaseConverter, allow it to initialize
    if isinstance(converter, BaseConverter):
        converter._delayed_init()
    # return
    return converter


def set_default_converter(converter=DefaultConverter):
    '''
    Sets the default converter used for view function parameters.

    `converter` should be a callable or a class with a __call__ method.
    If None, the system returns to the default DMP converter.
    '''
    DMP_OPTIONS[DEFAULT_CONVERTER_KEY] = _check_converter(converter)


def get_default_converter():
    '''
    Returns the default converter in the DMP system.
    '''
    return DMP_OPTIONS[DEFAULT_CONVERTER_KEY]


# set the intiial converter for the system
set_default_converter()


#!/usr/bin/env python3
'''
Created by Conan Albrecht <doconix@gmail.com>
Apache open source license.
November, 2017
'''
##################################################
###  Unique id generator.  Similar to uuid1() but 
###  also includes the process id.
###
###  Note that upping the counter requires a global lock.
###
###  The bit assignment:
###  
###      52 bits for nanoseconds since epoch (really it can use unlimited bits because on left side of the number, but 52 bits gets us to ~2100)
###      16 bits for counter
###      48 bits for machine id
###      24 bits for process id
###   ========
###     140 bits total, or 35 hex characters 
###  
###  Maximum number is 1.39e42
###

import uuid
import time as time  
import os
import random
import threading
import math
import collections
    
    
# initial values/constants
lastnow = 0
counterstart = random.getrandbits(16) - 1
countermax = math.pow(2, 16) - 1
counter = counterstart
# returns a 48 bit number
machineid = uuid.getnode()
# linux is usually 16 bits
processid = os.getpid()

# the main data structure
UID = collections.namedtuple('UID', ( 'time', 'counter', 'machine', 'process' ))

# binary size of each number
# and binary positions for shifting
# and for splitting hex and binary (from right side)
size = UID(52, 16, 48, 24)
_shift = []
for i in reversed(range(len(size))):
    _shift.append(sum(size[i:]))
shift = UID(*reversed(_shift))
hsplit = UID(*(int(s/-4) for s in shift))
bsplit = UID(*(s*-1 for s in shift))

######################################
###  Main API

def ruid():
    '''
    Creates a "raw" unique id.  The result is a
    UID namedtuple with four parts:
    
        time
        counter
        machine
        process
    
    All other functions in this module just format
    the id created in this function.
    '''
    global lastnow, counter_start, counter
    # update the nanoseconds and counter
    with threading.RLock():
        now = int(time.time())#int(time.time() * 1e6)
        counter += 1
        if counter >= countermax:
            counter = 0
        while now == lastnow and counter == counterstart:
            time.sleep(.001)  # wait a millisecond and try again
            now = int(time.time())#int(time.time() * 1e6)
        lastnow = now
    # return the named tuple
    return UID(now, counter, machineid, processid)
    

def iuid(raw=None):
    '''
    Creates a unique id as an int. 

    If provided, raw should be a UID named tuple
    (usually from a call to ruid).
    '''    
    if raw is None:
        raw = ruid()
    return (raw.time << shift.counter) + \
           (raw.counter << shift.machine) + \
           (raw.machine << shift.process) + \
           (raw.process)
    

def uid(raw=None, sep=None):
    '''
    Creates a unique id as a hex string. 

    If provided, raw should be a UID named tuple
    (usually from a call to ruid).
    
    Use sep='-' to separate the parts by dashes.
    '''
    if raw is None:
        raw = ruid()
    # hex version
    if sep is None:
        return '{:0x}'.format(iuid(raw))
    # pretty version
    n = uid(raw)
    return sep.join((
        n[:hsplit.counter], 
        n[hsplit.counter: hsplit.machine], 
        n[hsplit.machine: hsplit.process], 
        n[hsplit.process:],
    ))
    
    
def buid(raw=None, sep=None):
    '''
    Creates a unique id as a binary string. 

    If provided, raw should be a UID named tuple
    (usually from a call to ruid).
    
    Use sep='-' to separate the parts by dashes.
    '''
    if raw is None:
        raw = ruid()
    # hex version
    if sep is None:
        return '{:0b}'.format(iuid(raw))
    # pretty version
    n = buid(raw)
    return sep.join((
        n[:bsplit.counter], 
        n[bsplit.counter: bsplit.machine], 
        n[bsplit.machine: bsplit.process], 
        n[bsplit.process:],
    ))
    

def wuid(raw=None, leading='u'):
    '''
    Creates a unique id as a web-compliant id
    for use in HTML ids.  This is the same as 
    a hex id, but it has a leading `u` to ensure
    an alphabetical character comes first, per 
    the standard.

    If provided, raw should be a UID named tuple
    (usually from a call to ruid).

    Use sep='-' to separate the parts by dashes.
    '''
    if raw is None:
        raw = ruid()
    return '{}{}'.format(leading, uid(raw))
    

def iunpack(n):
    '''
    Unpacks the given integer number
    into a UID namedtuple.
    '''
    # format of these is (mask & n) >> shifted
    return UID(
        n >> shift.counter,
        ((((1 << size.counter) - 1) << shift.machine) & n) >> shift.machine,
        ((((1 << size.machine) - 1) << shift.process) & n) >> shift.process,
          ((1 << shift.process) - 1) & n,
    )
    
 
def unpack(hex_n):
    '''
    Unpacks the given hex number string
    into a UID namedtuple.
    
    To unpack a web id, use
        unpack(myid[1:])
    to remove the leading character.
    '''
    return iunpack(int(hex_n, 16))
    
    
    
    
###################################################
###  Unit tests for this module:
###
###     python3 uid.py
###

import unittest

class Tester(unittest.TestCase):

    def test_ruid(self):
        u = ruid()
        u2 = ruid()
        self.assertEqual(u.machine, u2.machine)
        self.assertEqual(u.process, u2.process)

    def test_int_hex_binary(self):
        u = ruid()
        n = iuid(u)
        h = uid(u)
        b = buid(u)
        self.assertEqual(n, int(h, 16))
        self.assertEqual(n, int(b, 2))
        
    def test_int_hex_binary(self):
        u = ruid()
        n = iuid(u)
        h = uid(u)
        b = buid(u)
        self.assertEqual(n, int(h, 16))
        self.assertEqual(n, int(b, 2))
        
    def test_pretty(self):
        u = ruid()
        # hex
        h = uid(u)
        p = uid(u, '-')
        self.assertEqual(h, p.replace('-', ''))
        # binary
        b = buid(u)
        p = buid(u, '-')
        self.assertEqual(b, p.replace('-', ''))

    def test_unpack(self):
        # one test
        u = ruid()
        self.assertEqual(u, unpack(uid(u)))
        self.assertEqual(u, iunpack(iuid(u)))
        # other direction with int
        n = iuid()
        self.assertEqual(n, iuid(iunpack(n)))
        # other direction with hex
        h = uid()
        self.assertEqual(h, uid(unpack(h)))
        
if __name__ == '__main__':
    unittest.main()
        


from django.conf import settings

from ..util import merge_dicts, getdefaultattr, flatten
from .static_links import LinkProvider

import os
import os.path
import posixpath


EMPTY_SET = set()


class AppJsBundleProvider(LinkProvider):
    '''
    Generates a JS <script> for the app-level bundle.
    '''
    default_options = merge_dicts(LinkProvider.default_options, {
        'group': 'scripts',
        'filename': lambda pr: os.path.join(*flatten(pr.app_config.path, 'scripts', '__bundle__.js')),
        'async': False,
    })

    def start(self, provider_run, data):
        data['templates'] = []

    def provide(self, provider_run, data):
        # instead of adding a script tag for each template (like the JsLinkProider does),
        # we just need to trigger the function for this page from the function
        data['templates'].append(posixpath.join(*flatten(self.app_config.name, self.subdir_parts[1:], self.template_name)))

    def finish(self, provider_run, data):
        '''Runs only on the main template in the chain.'''
        if self.mtime == 0 or len(data['templates']) == 0:
            return
        # add a <script> tag for the bundle (unless we already did so on this request)
        previous_bundles = getdefaultattr(provider_run.request.dmp, '__prev_app_bundles__', factory=set) if provider_run.request is not None else EMPTY_SET
        if self.filename not in previous_bundles:
            provider_run.write('<script data-context="{contextid}" src="{static}{href}?{version}"{async}></script>'.format(
                contextid=provider_run.uid,
                static=settings.STATIC_URL,
                href=self.filename,
                version=provider_run.version_id if provider_run.version_id is not None else self.mtime,
                async=' async="async"' if self.options['async'] else '',
            ))
            previous_bundles.add(self.filename)
        # trigger the functions for the template chain
        provider_run.write('<script data-context="{}">'.format(provider_run.uid))
        for template in data['templates']:
            provider_run.write('if (DMP_CONTEXT.appBundles["%s"]) { DMP_CONTEXT.appBundles["%s"]() };' % (template, template))
        provider_run.write('</script>')






from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from .base import BaseProvider
from ..util import DMP_OPTIONS, split_app, merge_dicts
from ..uid import wuid

import io


##################################################
###   Static File Provider Factory


def init_providers():
    '''Called when the DMP template engine is created by Django'''
    DMP_OPTIONS['RUNTIME_PROVIDER_FACTS'] = create_factories()


def create_factories(key='CONTENT_PROVIDERS'):
    '''Called from here as well as dmp_webpack.py'''
    providers = ( ProviderFactory(provider_def) for provider_def in DMP_OPTIONS[key] )
    return [ pf for pf in providers if pf.options['enabled'] ]


class ProviderFactory(object):
    '''Creator for a given Provider definition in settings.py.'''
    def __init__(self, provider_def):
        self.options = {}
        try:
            self.provider_class = provider_def['provider']
        except KeyError:
            raise ImproperlyConfigured('The Django Mako Plus template OPTIONS were not set up correctly in settings.py; a CONTENT_PROVIDERS item is missing `provider`.')
        if isinstance(self.provider_class, str):
            self.provider_class = import_string(self.provider_class)
        if not issubclass(self.provider_class, BaseProvider):
            raise ImproperlyConfigured('The Django Mako Plus template OPTIONS were not set up correctly in settings.py; The `provider` value must be a subclass of django_mako_plus.BaseProvider.')
        self.options = merge_dicts(self.provider_class.default_options, provider_def)

    def create(self, app_config, template_file):
        return self.provider_class(app_config, template_file, self.options)




####################################################
###   Main runner

# key to attach Provider objects to Mako Template objects during producting mode.
# I attach it to template objects because templates are already cached by mako, so
# caching them here would result in double-caching.
PROVIDERS_KEY = '_django_mako_plus_providers_'


class ProviderRun(object):
    '''A run through a template inheritance'''
    def __init__(self, tself, version_id=None, group=None, factories=None):
        self.uid = wuid()                           # a unique id for this run
        self.template = tself.template              # the template object
        self.request = tself.context.get('request') # request from the render()
        self.context = tself.context                # context from the render()
        self.group = group                          # the provider group being rendered (usually None)
        self.version_id = version_id                # unique number for overriding the cache (see LinkProvider)
        self._buffer = None                         # html to send back to the browser (providers add to this)
        # set up the factories we'll use
        if factories is None:
            factories = [
                pf for pf in DMP_OPTIONS['RUNTIME_PROVIDER_FACTS']
                if group is None or pf.options['group'] == group
            ]
        # provider data is separate from provider objects because a template (and its providers) can be in many chains at once,
        # such as base.htm's providers being in almost every chain on the site. this makes it unique to each provider on this run.
        self.provider_data = [ {} for i in range(len(factories)) ]
        # discover the ancestor templates for this template `self`
        self.chain = []
        # a set of providers for each template
        while tself is not None:
            # check if already attached to template, create if necessary
            providers = getattr(tself.template, PROVIDERS_KEY, None)
            if providers is None:
                app_config, template_file = split_app(tself.template.filename)
                if app_config is None:
                    raise ImproperlyConfigured("Could not determine the app for template {}. Is this template's app registered as a DMP app?".format(tself.template.filename))
                providers = [ pf.create(app_config, template_file) for pf in factories ]
                if not settings.DEBUG: # attach to template for speed in production mode
                    setattr(tself.template, PROVIDERS_KEY, providers)
            self.chain.append(providers)
            # loop with the next inherited template
            tself = tself.inherits
        # we need furthest ancestor first, current template last so current template (such as CSS) can override ancestors
        self.chain.reverse()

    def get_content(self):
        '''Runs the providers for each template in the current inheritance chain, returning the combined content.'''
        if len(self.chain) == 0:
            return ''
        self._buffer = io.StringIO()
        # start() on the last provider list in the chain
        for provider, data in zip(self.chain[-1], self.provider_data):
            provider.start(self, data)
        # provide() on the all provider lists in the chain
        for template_provider_list in self.chain:
            for provider, data in zip(template_provider_list, self.provider_data):
                provider.provide(self, data)
        # finish() on the last provider list in the chain
        for provider, data in zip(self.chain[-1], self.provider_data):
            provider.finish(self, data)
        return self._buffer.getvalue()

    def write(self, content):
        self._buffer.write(content)
        if settings.DEBUG:
            self._buffer.write('\n')



from django import template
from django.utils.safestring import mark_safe
from ..convenience import render_template

register = template.Library()

@register.simple_tag(takes_context=True)
def dmp_include(context, app, template_name, def_name=None):
    '''Standard Django tag to include a DMP (mako) template.'''
    return mark_safe(render_template(
        request=context.get('request'), 
        app=app, 
        template_name=template_name, 
        context=context.flatten(), 
        def_name=def_name
    ))



from django.conf import settings
from django.http import HttpResponse
from django.template import TemplateDoesNotExist, TemplateSyntaxError, Context, RequestContext

from mako.exceptions import TopLevelLookupException, TemplateLookupException, CompileException, SyntaxException, html_error_template
from mako.lookup import TemplateLookup
from mako.template import Template

from .exceptions import RedirectException
from .signals import dmp_signal_pre_render_template, dmp_signal_post_render_template, dmp_signal_redirect_exception
from .util import get_dmp_instance, log, DMP_OPTIONS

import logging
import mimetypes
import os
import os.path
import sys



class DMPTemplateLookup(TemplateLookup):
    '''Small extension to Mako's template lookup to provide a link back to the MakoTemplateLoader'''
    def __init__(self, template_loader, *args, **kwargs):
        super(DMPTemplateLookup, self).__init__(*args, **kwargs)
        self.template_loader = template_loader


class MakoTemplateLoader(object):
    '''Finds Mako templates for a Django app.'''
    def __init__(self, app_path, template_subdir='templates'):
        '''
        The loader looks in the app_path/templates directory unless
        the template_subdir parameter overrides this default.

        You should not normally create this object because it bypasses
        the DMP cache.  Instead, call get_template_loader() or
        get_template_loader_for_path().
        '''
        self.app_path = app_path
        # calculate the template directory and check that it exists
        if template_subdir is None:  # None skips adding the template_subdir
            self.template_dir = os.path.abspath(app_path)
        else:
            self.template_dir = os.path.abspath(os.path.join(app_path, template_subdir))
        # I used to check for the existence of the template dir here, but it caused error
        # checking at engine load time (too soon).  I now wait until get_template() is called,
        # which fails with a TemplateDoesNotExist exception if the template_dir doesn't exist.

        # calculate the cache root and template search directories
        self.cache_root = os.path.join(self.template_dir, DMP_OPTIONS['TEMPLATES_CACHE_DIR'])
        self.template_search_dirs = [ self.template_dir ]
        self.template_search_dirs.extend(DMP_OPTIONS['TEMPLATES_DIRS'])
        # Mako doesn't allow parent directory inheritance, such as <%inherit file="../../otherapp/templates/base.html"/>
        # including the project base directory allows this through "absolute" like <%inherit file="/otherapp/templates/base.html"/>
        # (note the leading slash, which means BASE_DIR)
        self.template_search_dirs.append(settings.BASE_DIR)

        # create the actual Mako TemplateLookup, which does the actual work
        self.tlookup = DMPTemplateLookup(self, directories=self.template_search_dirs, imports=DMP_OPTIONS['RUNTIME_TEMPLATE_IMPORTS'], module_directory=self.cache_root, collection_size=2000, filesystem_checks=settings.DEBUG, input_encoding=DMP_OPTIONS['DEFAULT_TEMPLATE_ENCODING'])


    def get_template(self, template, def_name=None):
        '''Retrieve a *Django* API template object for the given template name, using the app_path and template_subdir
           settings in this object.  This method still uses the corresponding Mako template and engine, but it
           gives a Django API wrapper around it so you can use it the same as any Django template.

           If def_name is provided, template rendering will be limited to the named def/block (see Mako docs).

           This method corresponds to the Django templating system API.
           A Django exception is raised if the template is not found or cannot compile.
        '''
        try:
            # wrap the mako template in an adapter that gives the Django template API
            return MakoTemplateAdapter(self.get_mako_template(template), def_name)
        except (TopLevelLookupException, TemplateLookupException) as e: # Mako exception raised
            raise TemplateDoesNotExist('Template "%s" not found in search path: %s.' % (template, self.template_search_dirs))
        except (CompileException, SyntaxException) as e: # Mako exception raised
            raise TemplateSyntaxError('Template "%s" raised an error: %s' % (template, e))


    def get_mako_template(self, template, force=False):
        '''Retrieve the real *Mako* template object for the given template name without any wrapper,
           using the app_path and template_subdir settings in this object.

           This method is an alternative to get_template().  Use it when you need the actual Mako template object.
           This method raises a Mako exception if the template is not found or cannot compile.

           If force is True, an empty Mako template will be created when the file does not exist.
           This option is used by the providers part of DMP and normally be left False.
        '''
        if template is None:
            raise TemplateLookupException('Template "%s" not found in search path: %s.' % (template, self.template_search_dirs))
        # get the template
        try:
            template_obj = self.tlookup.get_template(template)
        except TemplateLookupException:
            if not force:
                raise
            template_obj = Template('', filename=os.path.join(self.template_dir, template))

        # get the template
        return template_obj




class MakoTemplateAdapter(object):
    '''A thin wrapper for a Mako template object that provides the Django API methods.'''
    def __init__(self, mako_template, def_name=None):
        '''
        Creates an adapter that corresponds to the Django API.

        If def_name is provided, template rendering will be limited to the named def/block (see Mako docs).
        This can also be provided in the call to render().
        '''
        self.mako_template = mako_template
        self.def_name = def_name


    @property
    def engine(self):
        '''Returns the DMP engine (method required by Django specs)'''
        return get_dmp_instance()


    def render(self, context=None, request=None, def_name=None):
        '''
        Renders a template using the Mako system.  This method signature conforms to
        the Django template API, which specifies that template.render() returns a string.

            @context  A dictionary of name=value variables to send to the template page.  This can be a real dictionary
                      or a Django Context object.
            @request  The request context from Django.  If this is None, any TEMPLATE_CONTEXT_PROCESSORS defined in your settings
                      file will be ignored but the template will otherwise render fine.
            @def_name Limits output to a specific top-level Mako <%block> or <%def> section within the template.
                      If the section is a <%def>, any parameters must be in the context dictionary.  For example,
                      def_name="foo" will call <%block name="foo"></%block> or <%def name="foo()"></def> within
                      the template.

        Returns the rendered template as a unicode string.

        The method triggers two signals:
            1. dmp_signal_pre_render_template: you can (optionally) return a new Mako Template object from a receiver to replace
               the normal template object that is used for the render operation.
            2. dmp_signal_post_render_template: you can (optionally) return a string to replace the string from the normal
               template object render.
        '''
        # set up the context dictionary, which is the variables available throughout the template
        context_dict = {}
        # if request is None, add some default items because the context processors won't happen
        if request is None:
            context_dict['settings'] = settings
            context_dict['STATIC_URL'] = settings.STATIC_URL
        # let the context_processors add variables to the context.
        if not isinstance(context, Context):
            context = Context(context) if request is None else RequestContext(request, context)
        with context.bind_template(self):
            for d in context:
                context_dict.update(d)
        context_dict.pop('self', None)  # some contexts have self in them, and it messes up render_unicode below because we get two selfs

        # send the pre-render signal
        if DMP_OPTIONS['SIGNALS'] and request is not None:
            for receiver, ret_template_obj in dmp_signal_pre_render_template.send(sender=self, request=request, context=context, template=self.mako_template):
                if ret_template_obj is not None:
                    if isinstance(ret_template_obj, MakoTemplateAdapter):
                        self.mako_template = ret_template_obj.mako_template   # if the signal function sends a MakoTemplateAdapter back, use the real mako template inside of it
                    else:
                        self.mako_template = ret_template_obj                 # if something else, we assume it is a mako.template.Template, so use it as the template

        # do we need to limit down to a specific def?
        # this only finds within the exact template (won't go up the inheritance tree)
        # I wish I could make it do so, but can't figure this out
        render_obj = self.mako_template
        if def_name is None:
            def_name = self.def_name
        if def_name:  # do we need to limit to just a def?
            render_obj = self.mako_template.get_def(def_name)

        # PRIMARY FUNCTION: render the template
        template_name = '%s::%s' % (self.mako_template.filename or 'string', def_name or 'body')
        if log.isEnabledFor(logging.INFO):
            log.info('rendering template %s', template_name)
        if settings.DEBUG:
            try:
                content = render_obj.render_unicode(**context_dict)
            except Exception as e:
                log.exception('exception raised during template rendering: %s', e)  # to the console
                content = html_error_template().render_unicode()       # to the browser
        else:  # this is outside the above "try" loop because in non-DEBUG mode, we want to let the exception throw out of here (without having to re-raise it)
            content = render_obj.render_unicode(**context_dict)

        # send the post-render signal
        if DMP_OPTIONS['SIGNALS'] and request is not None:
            for receiver, ret_content in dmp_signal_post_render_template.send(sender=self, request=request, context=context, template=self.mako_template, content=content):
                if ret_content is not None:
                    content = ret_content  # sets it to the last non-None return in the signal receiver chain

        # return
        return content


    def render_to_response(self, context=None, request=None, def_name=None, content_type=None, status=None, charset=None):
        '''
        Renders the template and returns an HttpRequest object containing its content.

        This method returns a django.http.Http404 exception if the template is not found.
        If the template raises a django_mako_plus.RedirectException, the browser is redirected to
        the given page, and a new request from the browser restarts the entire DMP routing process.
        If the template raises a django_mako_plus.InternalRedirectException, the entire DMP
        routing process is restarted internally (the browser doesn't see the redirect).

            @request      The request context from Django.  If this is None, any TEMPLATE_CONTEXT_PROCESSORS defined in your settings
                          file will be ignored but the template will otherwise render fine.
            @template     The template file path to render.  This is relative to the app_path/controller_TEMPLATES_DIR/ directory.
                          For example, to render app_path/templates/page1, set template="page1.html", assuming you have
                          set up the variables as described in the documentation above.
            @context      A dictionary of name=value variables to send to the template page.  This can be a real dictionary
                          or a Django Context object.
            @def_name     Limits output to a specific top-level Mako <%block> or <%def> section within the template.
                          For example, def_name="foo" will call <%block name="foo"></%block> or <%def name="foo()"></def> within the template.
            @content_type The MIME type of the response.  Defaults to settings.DEFAULT_CONTENT_TYPE (usually 'text/html').
            @status       The HTTP response status code.  Defaults to 200 (OK).
            @charset      The charset to encode the processed template string (the output) with.  Defaults to settings.DEFAULT_CHARSET (usually 'utf-8').

        The method triggers two signals:
            1. dmp_signal_pre_render_template: you can (optionally) return a new Mako Template object from a receiver to replace
               the normal template object that is used for the render operation.
            2. dmp_signal_post_render_template: you can (optionally) return a string to replace the string from the normal
               template object render.
        '''
        try:
            if content_type is None:
                content_type = mimetypes.types_map.get(os.path.splitext(self.mako_template.filename)[1].lower(), settings.DEFAULT_CONTENT_TYPE)
            if charset is None:
                charset = settings.DEFAULT_CHARSET
            if status is None:
                status = 200
            content = self.render(context=context, request=request, def_name=def_name)
            return HttpResponse(content.encode(charset), content_type='%s; charset=%s' % (content_type, charset), status=status)

        except RedirectException: # redirect to another page
            e = sys.exc_info()[1]
            if request is None:
                log.info('a template redirected processing to %s', request.dmp.module, request.dmp.function, e.redirect_to)
            else:
                log.info('view function %s.%s redirected processing to %s', request.dmp.module, request.dmp.function, e.redirect_to)
            # send the signal
            if DMP_OPTIONS['SIGNALS']:
                dmp_signal_redirect_exception.send(sender=sys.modules[__name__], request=request, exc=e)
            # send the browser the redirect command
            return e.get_response(request)








from django.conf import settings
from django.utils.module_loading import import_string

from ..util import merge_dicts, flatten, log
from ..version import __version__
from .base import BaseProvider

import logging
import os
import os.path
import json
import glob
from collections import namedtuple



class LinkProvider(BaseProvider):
    '''
    Renders links like <link> and <script> based on the name of the template
    and supertemplates.
    '''
    # the default options are for CSS files
    default_options = merge_dicts(BaseProvider.default_options, {
        'group': 'static.file',
        # function that returns the filename that should be checked and if exists, linked by the provider
        'filename': lambda pr: 'app/subdir/template.ext',
    })
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        filepath = self.options['filename'](self)
        self.filename = os.path.relpath(filepath, settings.BASE_DIR).replace('\\', '/')
        if log.isEnabledFor(logging.DEBUG):
            log.debug('[%s] %s searching for %s', self.template_file, self.__class__.__qualname__, self.filename)
        try:
            self.mtime = int(os.stat(filepath).st_mtime)
            if log.isEnabledFor(logging.INFO):
                log.info('[%s] found %s', self.template_file, self.filename)
        except FileNotFoundError:
            self.mtime = 0


class CssLinkProvider(LinkProvider):
    '''Generates a CSS <link>'''
    default_options = merge_dicts(LinkProvider.default_options, {
        'group': 'styles',
        'filename': lambda pr: os.path.join(*flatten(pr.app_config.path, 'styles', pr.subdir_parts[1:], pr.template_name + '.css')),
        'skip_duplicates': True,
    })

    def provide(self, provider_run, data):
        if self.mtime == 0:
            return
        provider_run.write('<link data-context="{contextid}" rel="stylesheet" type="text/css" href="{static}{url}?{version}" />'.format(
            contextid=provider_run.uid,
            static=settings.STATIC_URL,
            url=self.filename,
            version=provider_run.version_id if provider_run.version_id is not None else self.mtime,
            skip_duplicates='true' if self.options['skip_duplicates'] else 'false',
        ))


class JsLinkProvider(LinkProvider):
    '''Generates a JS <script>.'''
    default_options = merge_dicts(LinkProvider.default_options, {
        'group': 'scripts',
        'filename': lambda pr: os.path.join(*flatten(pr.app_config.path, 'scripts', pr.subdir_parts[1:], pr.template_name + '.js')),
        'async': False,
    })

    def start(self, provider_run, data):
        data['urls'] = []

    def provide(self, provider_run, data):
        # delaying printing of tag because the JsContextProvider delays and this must go after it
        if self.mtime == 0:
            return
        data['urls'].append('{}?{}'.format(
            self.filename.replace('\\', '/'),
            provider_run.version_id if provider_run.version_id is not None else self.mtime,
        ))

    def finish(self, provider_run, data):
        for url in data['urls']:
            provider_run.write('<script data-context="{contextid}" src="{static}{url}"{async}></script>'.format(
                contextid=provider_run.uid,
                static=settings.STATIC_URL,
                url=url,
                async=' async="async"' if self.options['async'] else '',
            ))


class JsContextProvider(BaseProvider):
    '''
    Adds all js_context() variables to DMP_CONTEXT.
    This should be listed before JsLinkProvider so the
    context variables are available during <script> runs.
    '''
    default_options = merge_dicts(LinkProvider.default_options, {
        'group': 'scripts',
        'encoder': 'django.core.serializers.json.DjangoJSONEncoder',
    })

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.encoder = import_string(self.options['encoder'])
        self.template = "{}/{}".format(self.app_config.name, self.template_name)

    def start(self, provider_run, data):
        data['templates'] = []

    def provide(self, provider_run, data):
        data['templates'].append(self.template)

    def finish(self, provider_run, data):
        if len(data['templates']) == 0:
            return
        context_data = {
            jscontext('__router__'): {
                'template': self.template,
                'app': provider_run.request.dmp.app if provider_run.request is not None else None,
                'page': provider_run.request.dmp.page if provider_run.request is not None else None,
            },
        }
        for k in provider_run.context.kwargs:
            if isinstance(k, jscontext):
                value = provider_run.context[k]
                context_data[k] = value.__jscontext__() if hasattr(value, '__jscontext__') else value
        provider_run.write('<script>')
        provider_run.write('DMP_CONTEXT.set("{version}", "{contextid}", {data}, {templates});'.format(
            version=__version__,
            contextid=provider_run.uid,
            data=json.dumps(context_data, cls=self.encoder, separators=(',', ':')) if context_data else '{}',
            templates=json.dumps(data['templates']),
        ))
        provider_run.write('</script>')


class jscontext(str):
    '''
    Marks a key in the context dictionary as a JS context item.
    JS context items are sent to the template like normal,
    but they are also added to the runtime JS namespace.

    See the tutorial for more information on this function.
    '''
    # no code needed, just using the class for identity



from django.template import RequestContext, engines

from .util import log




############################################################################
###   Rendering template blocks in other syntax languages

def django_syntax(local, **kwargs):
    '''
    A Mako filter that renders a block of text using the standard Django template engine.
    The Django template engine must be listed in settings.TEMPLATES.

    The template context variables are available in the embedded template.
    Specify kwargs to add additional variables created within the template.

    ## Simple expression in Django syntax:
    ${ '{{ name }}' | django_syntax(local) }

    ## Embedded code block:
    <%block filter="django_syntax(local)">
        {% for story in story_list %}
            <h2>
                {{ story.headline|upper }}
            </h2>
        {% endfor %}
    </%block>

    '''
    return alternate_syntax(local, 'django', **kwargs)


def jinja2_syntax(local, **kwargs):
    '''
    A Mako filter that renders a block of text using the Jinja2 template engine.
    The Jinja2 template engine must be listed in settings.TEMPLATES.

    The template context variables are available in the embedded template.
    Specify kwargs to add additional variables created within the template.

    ## Simple expression in Jinja2 syntax:
    ${ '{{ name }}' | jinja2_syntax(local) }

    ## Embedded Jinja2 code block:
    <%block filter="jinja2_syntax(local)">
        {% for item in navigation %}
            <li><a href="{{ item.href }}">{{ item.caption }}</a></li>
        {% endfor %}
    </%block>

    '''
    return alternate_syntax(local, 'jinja2', **kwargs)


def alternate_syntax(local, using, **kwargs):
    '''
    A Mako filter that renders a block of text using a different template engine
    than Mako.  The named template engine must be listed in settings.TEMPLATES.

    The template context variables are available in the embedded template.
    Specify kwargs to add additional variables created within the template.

    The following examples assume you have installed the django_mustache template
    engine in settings.py:

        ## Simple expression in Mustache syntax:
        ${ '{{ name }}' | template_syntax(local, 'django_mustache') }

        ## Embedded Mustache code block:
        <%block filter="template_syntax(local, 'django_mustache')">
            {{#repo}}
                <b>{{name}}</b>
            {{/repo}}
            {{^repo}}
                No repos :(
            {{/repo}}
        </%block>

    Rendering Django or Jinja2 templates should be done with `django_syntax` and
    `jinja2_syntax` because it doesn't require the using parameter.
    '''
    # get the request (the MakoTemplateAdapter above places this in the context)
    request = local.context['request'] if isinstance(local.context, RequestContext) else None
    # get the current Mako template object so we can attach the compiled string for later use
    # Mako caches and automatically recreates this if the file changes
    mako_template = local.template
    if not hasattr(mako_template, '__compiled_template_syntax'):
        mako_template.__compiled_template_syntax = {}

    # create a closure so we can still get to context and using (Mako filters take exactly one parameter: the string to filter)
    def wrap(template_st):
        # get the template object, or create and cache it
        try:
            template = mako_template.__compiled_template_syntax[template_st]
        except KeyError:
            engine = engines[using]
            template = engine.from_string(template_st)
            # using full string, even if long, as the key doesn't really affect performance of python's hash (see http://stackoverflow.com/questions/28150047/efficiency-of-long-str-keys-in-python-dictionary)
            mako_template.__compiled_template_syntax[template_st] = template

        # create a copy the context and add any kwargs to it
        dcontext = dict(local.context)
        dcontext.update(kwargs)

        # print a debug statement to the log
        log.debug('rendering embedded expression or block using %s template engine', using)

        # render the template with the context
        return template.render(context=dcontext, request=request)

    # return the embedded function
    return wrap

